//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Declares the DistToolBase class.  Distribution tool
//                  buffers and distributes circuit blocks (and related data
//                  such as option blocks, metadata, etc) for/during parsing.
//
// Special Notes  :
//
// Creator        : Eric Rankin, SNL
//
// Creation Date  : 03/12/2003
//
//
//
//
//-----------------------------------------------------------------------------


#ifndef Xyce_N_IO_DistToolBase_h
#define Xyce_N_IO_DistToolBase_h

#include <vector>
#include <string>

#include <N_IO_fwd.h>
#include <N_PDS_fwd.h>
#include <N_IO_CircuitContext.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_IO_DistributionTool.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Class          : DistToolBase
// Purpose        : Provide basic common functionality for all distribution tools.
//-----------------------------------------------------------------------------
class DistToolBase
{
public:
  DistToolBase(
    CircuitBlock &                           circuit_block,
    std::map<std::string,FileSSFPair>      & ssfMap );

  virtual ~DistToolBase() {}

protected:

  // Parse the given include file for 2nd pass
  virtual bool parseIncludeFile(std::string const& includeFiles,
                                const std::string &libSelect) = 0;

  // Expand a subcircuit instance by adding the devices and
  // device models that compose the subcircuit to the main
  // (top level) circuit. Prepend device names and nodes with
  // subcircuitPrefix.
  virtual bool expandSubcircuitInstance(DeviceBlock & subcircuitInstance,
                                        const std::string &libSelect, 
                                        std::string &libInside) = 0;

  // get the next line from the SpiceSeparatedField pointer
  bool getLine(TokenVector& line,
               const std::string &libSelect, std::string &libInside);

  // Process a device line on processor zero, or serial.
  bool handleDeviceLine(TokenVector const& deviceLine,
                        const std::string &libSelect, std::string &libInside);

  // Post process a mutual inductor in the current circuit.
  bool handleMutualInductance( DeviceBlock & device );

  // Fully parse and instantiate a single device.
  bool instantiateDevice( DeviceBlock & device,
      const std::string & prefix, const unordered_map<std::string,std::string>& nodeMap,
      const std::string &libSelect, std::string &libInside);

  // global data
  CircuitBlock &                circuitBlock_;
  CircuitContext *              circuitContext_;
  std::map<std::string,FileSSFPair>  & ssfMap_;
  std::string                   netlistFilename_;

  DeviceBlock                   device_;
  CircuitBlock*                 mainCircuitPtr_;
  CircuitBlock*                 parentCircuitPtr_;
  CircuitBlock*                 currentCircuitPtr_;
  std::vector<bool>             preprocessFilter_;
  SpiceSeparatedFieldTool*      ssfPtr_;

  bool                          remove_any_redundant_;  ///< check device nodes for removal
};

} // namespace IO
} // namespace Xyce

#endif // Xyce_N_IO_DistToolBase_h
