//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Declares the DistToolFlatRoundRobin class.  Distribution tool
//                  buffers and distributes circuit blocks (and related data
//                  such as option blocks, metadata, etc) for/during parsing.
//
// Special Notes  :
//
// Creator        : Eric Rankin, SNL
//
// Creation Date  : 03/12/2003
//
//
//
//
//-----------------------------------------------------------------------------


#ifndef Xyce_N_IO_DistToolFlatRoundRobin_h
#define Xyce_N_IO_DistToolFlatRoundRobin_h

#include <vector>
#include <string>

#include <N_IO_fwd.h>
#include <N_PDS_fwd.h>
#include <N_IO_CircuitContext.h>
#include <N_IO_SpiceSeparatedFieldTool.h>
#include <N_IO_DistributionTool.h>
#include <N_IO_DistToolBase.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Class          : DistToolFlatRoundRobin
// Purpose        : Buffers and distributes circuit blocks, device lines, and related data
//                  for/during parsing.  This class has the strategy of first-come-first-served
//                  distribution of devices over processors.
//-----------------------------------------------------------------------------
class DistToolFlatRoundRobin: public DistributionTool, public DistToolBase
{
public:
  DistToolFlatRoundRobin(
    Parallel::Communicator *                 pdsCommPtr,
    PkgOptionsMgr &                          options_manager, 
    CircuitBlock &                           circuit_block,
    std::map<std::string,FileSSFPair>      & ssfMap, 
    std::map<std::string, IncludeFileInfo> & iflMap);

  ~DistToolFlatRoundRobin();

  // send options, metatdata, and context to all procs
  bool broadcastGlobalData();

  // Distribute devices using hierarchical context object.
  void distributeDevices();

  // change current subcircuit context after a new subcircuit is started
  bool circuitStart( std::string const & subcircuitName,
                     std::vector<std::string> const & nodes,
                     std::string const & prefix,
                     std::vector<Device::Param> const & params )
  { // Throw error here!
    return false;
  }

  // change current subcircuit context to previous context
  bool circuitEnd()
  { // Throw error here!
    return false;
  }

protected:

  // Parse the given include file for 2nd pass
  bool parseIncludeFile(std::string const& includeFiles,
                        const std::string &libSelect);

  bool expandSubcircuitInstance(DeviceBlock & subcircuitInstance,
                                const std::string &libSelect,
                                std::string &libInside)
  { // Throw error here! 
    return false;
  }

private:

  // Send a circuit device line to current proc
  bool circuitDeviceLine(TokenVector & deviceLine );
  void endDeviceLines();

  void setFileName ( std::string const & fileNameIn );

  // send circuit context to all procs
  void setCircuitContext();

  // stage options for tx
  void setCircuitOptions();

  // receive all data from proc 0
  bool receiveCircuitData();

  // send buffer from proc 0
  void send(int size = -1);

  // process all the devices held in the buffer
  bool processDeviceBuffer();

private:

  void selectNextProcessor();

  Parallel::Communicator *      pdsCommPtr_;            ///< parallel services
  int                           numProcs_;              ///< total number of available procs for distribution
  int                           currProc_;              ///< proc that will receive the next tx

  // global data
  PkgOptionsMgr &               optionsManager_; 
  std::list<Util::OptionBlock> &options_;
  std::map<std::string, IncludeFileInfo> & iflMap_;

  int                           procDeviceCount_;       ///< max number of devices to tx to each proc
  int                           blockSize_;             ///< size of blocks used in round robin
  int                           numBlocks_;             ///< number of blocks used in round robin
  int                           devices_;               ///< total number of devices;
  int                           deviceLinesSent_;       ///< keep track of current txmitted devices
  int                           charBufferSize_;        ///< length of buffer
  int                           charBufferPos_;         ///< length of data packed in buffer

  std::vector<int>              totalDeviceLinesSent_;  ///< keep track of txmitted devices for all procs
  std::vector<char *>           bufs_;                  ///< tx/rx buffer
  std::vector<int>              bufSize_;               ///< buffer size
  char *                        charBuffer_;

  bool                          circuitContextReady_;   ///< flags for managing global data
  bool                          circuitOptionsReady_;   ///< flags for managing global data

};

} // namespace IO
} // namespace Xyce

#endif // Xyce_N_IO_DistToolFlatRoundRobin_h
