// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Defines the DistToolBase class.  Distribution tool
//                  buffers and distributes circuit blocks (and related data
//                  such as option blocks, metadata, etc) for/during parsing.
// Special Notes  :
//
// Creator        : Eric Rankin, SNL
//
// Creation Date  : 03/12/2003
//
//
//
//
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_IO_fwd.h>
#include <N_DEV_DeviceMgr.h>
#include <N_ERH_ErrorMgr.h>
#include <N_ERH_Message.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_DistToolBase.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_ParsingHelpers.h>
#include <N_PDS_Comm.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Stats.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Function      : DistToolBase::DistToolBase
// Purpose       : ctor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
DistToolBase::DistToolBase(
  CircuitBlock                           & circuit_block,
  std::map<std::string,FileSSFPair>      & ssfMap
  )
  : circuitBlock_(circuit_block),
    circuitContext_(circuit_block.getCircuitContextPtr()),
    ssfMap_(ssfMap),
    netlistFilename_(),
    device_(*circuit_block.getCircuitContextPtr(), circuit_block.getMetadata()),
    mainCircuitPtr_(&circuit_block), 
    parentCircuitPtr_(NULL),
    currentCircuitPtr_(&circuit_block),
    preprocessFilter_(PreprocessType::NUM_PREPROCESS, false),
    ssfPtr_(0),
    remove_any_redundant_(false)
{
}

//----------------------------------------------------------------------------
// Function       : DistToolBase::getLine
// Purpose        :
// Special Notes  :
// Scope          : private
// Creator        : Lon Waters
// Creation Date  : 08/01/2003
//----------------------------------------------------------------------------
bool DistToolBase::getLine( TokenVector &line,
  const std::string &libSelect, std::string &libInside)
{
  int eof = 0;

  while (!eof)
  {
    {
      //Stats::StatTop _getLineStat("Tokenize Line");
      //Stats::TimeBlock _getLineTimer(_getLineStat);

      eof = !ssfPtr_->getLine(line,preprocessFilter_[PreprocessType::REPLACE_GROUND]); // Breaks the line into fields.
    }

    if (DEBUG_IO) 
    {
      Xyce::dout() << "Pass 2 read netlist line:  "<< std::endl;
      for (unsigned int i = 0; i < line.size(); ++i)
      {
        Xyce::dout() << line[i].string_ << " ";
      }
      Xyce::dout() << std::endl;

      if (eof)
      {
        Xyce::dout() << "We are at the end of file, size of line is " << line.size()
                     << std::endl;
      }
    }

    // better not try to do anything if getLine returned an empty line!
    if ( line.empty() )
      continue;

    // Determine what to do with the parsed line.
    ExtendedString ES1(line[0].string_);
    ES1.toUpper();
    if ( !(libSelect != libInside && libInside != "" && ES1 != ".ENDL"))
    {
      char lineType;
      lineType = ES1[0];
      bool removecomponent=false;

      if (lineType >= 'A' && lineType <= 'Z')
      {

        if ( lineType == 'C' || lineType == 'D' || lineType == 'I' ||
             lineType == 'L' || lineType == 'R' || lineType == 'V')
        {
          if (remove_any_redundant_ && line.size() > 2) //make sure that there are two nodes to check!
          {
            ExtendedString node1 ( line[1].string_ );
            ExtendedString node2 ( line[2].string_ );
            node1.toUpper();
            node2.toUpper();

            removecomponent = IO::removeTwoTerminalDevice(preprocessFilter_,
                                                          lineType, node1, node2);
          }
        }
        else if (lineType == 'M' || lineType == 'Q')
        {
          if (remove_any_redundant_ && line.size() > 3) //make sure that there are three nodes to check!
          {
            ExtendedString node1 ( line[1].string_ );
            ExtendedString node2 ( line[2].string_ );
            ExtendedString node3 ( line[3].string_ );
            node1.toUpper();
            node2.toUpper();
            node3.toUpper();

            removecomponent = IO::removeThreeTerminalDevice(preprocessFilter_,
                                                            lineType, node1,
                                                            node2, node3);
          }
        }
        if (!removecomponent)
        {
          if (lineType != 'X' && lineType != 'K' && lineType != 'L')
          {
            // check for .INITCOND values
            std::string tmpName(circuitContext_->getPrefix());
            if (tmpName == "")
              tmpName = ES1;
            else
              tmpName = tmpName + ":" + ES1;

            if( mainCircuitPtr_->initCondIndex.find( tmpName ) !=
                mainCircuitPtr_->initCondIndex.end() )
            {
              // append the IC=val1...valN list to device line
              line.insert( line.end(),
                           ( mainCircuitPtr_->initCondIndex[tmpName] ).begin(),
                           ( mainCircuitPtr_->initCondIndex[tmpName] ).end() );
            }

            return true;
          }
          else if (lineType == 'L')
          {
            std::set< std::string > & cTable =
              circuitContext_->getAllCoupledInductors();

            if( cTable.find( ES1 ) == cTable.end() )
            {
              // treat as a normal device line if inductor is not coupled
              return true;
            }
            continue;
          }
          else if (lineType == 'X')
          {
            handleDeviceLine(line, libSelect, libInside);
          }
        }
        else
        {
          // This device is being removed, clear the line.
          line.clear();

          if (DEBUG_IO)
          {
            Xyce::dout() << "Netlist Parse 2:  ";
            Xyce::dout() << "removing component " << ES1 << ".  All nodes on the device";
            Xyce::dout() << " are the same."  << std::endl;
          }

          return true;
        }
      }
      else if (lineType == '.')
      {
        if (ES1 == ".SUBCKT")
        {
          // Jump to the end of the subcircuit.
          // Find the subcircuit corresponding to this instance.
          CircuitBlock* subcircuitPtr =
            currentCircuitPtr_->findSubcircuit(ExtendedString(line[1].string_).toUpper());

          ssfPtr_->setLocation(subcircuitPtr->getEndPosition());
          ssfPtr_->setLineNumber(subcircuitPtr->getLineEndPosition());

        }
        else if (ES1 == ".INCLUDE" || ES1 == ".INC" || ES1 == ".LIB")
        {
          std::string includeFile, libSelect_child;
          libSelect_child = libSelect;
          Xyce::IO::handleIncludeLine( (ssfMap_.size()==1), netlistFilename_, line,
                                       ES1, includeFile, libSelect_child, libInside );
          if (includeFile != "")
          {
            int result = this->parseIncludeFile(includeFile, libSelect_child);
            if (!result)
              return result;
          }
        }
        else if (ES1 == ".ENDL")
        {
          Xyce::IO::handleEndlLine ( netlistFilename_, line, libSelect, libInside );
        }
        else if (ES1 == ".ENDS")
        {
          return parentCircuitPtr_ == 0;
        }
        else if (ES1 == ".END")
        {
          return false;
        }
      }
      else
      {
        // Do nothing for other types of lines
      }
    }
  }
  // Only get here if end of file.

  return false;
}

//----------------------------------------------------------------------------
// Function       : DistToolBase::handleDeviceLine
// Purpose        : Processor zero / serial processing of devices
// Special Notes  :
// Scope          : public
// Creator        : Lon Waters
// Creation Date  : 02/21/2003
//----------------------------------------------------------------------------
bool DistToolBase::handleDeviceLine( TokenVector const& deviceLine,
    const std::string &libSelect, std::string &libInside)
{
  bool result;

  device_.clear();

  // Fully parse the device line and extract the relevant data.
  {
    //Stats::StatTop _extractData("Extract Device Data");
    //Stats::TimeBlock _extractDataTimer(_extractData);

    result = device_.extractData(netlistFilename_, deviceLine);
    if (!result)
      return result;
  }

  unordered_map<std::string, std::string> * nodeMapPtr = circuitContext_->getNodeMapPtr();
  result = instantiateDevice(device_, circuitContext_->getPrefix(), *nodeMapPtr, libSelect, libInside);

  return result;
}

//----------------------------------------------------------------------------
// Function       : DistToolBase::instantiateDevice
// Purpose        : Extract data from tokenized device and place all data about
//                  the device in Topology.
// Special Notes  : This method is misnamed as it does not actually instantiate
//                  the device.  That is done after partitioning (if parallel)
//                  and is called from Topology.
//                  This actually fills the DeviceBlock which is stored in
//                  Topology.
// Scope          : private
// Creator        : Lon Waters
// Creation Date  : 07/28/2003
//----------------------------------------------------------------------------
bool DistToolBase::instantiateDevice(
  DeviceBlock &                                 device,
  const std::string &                           prefix,
  const unordered_map<std::string,std::string> &nodeMap,
  const std::string &                           libSelect,
  std::string &                                 libInside)
{
  bool result = true;

  unordered_map<std::string,std::string>::const_iterator nodeMapIter;
  unordered_map<std::string,std::string>::const_iterator nodeMapEnd = nodeMap.end();

  // Global params were discovered by proc 0 in pass 1 and stored in root level of
  // circuitContext_->  Then circuitContext was transmitted to all procs.

  // Now, before device parsing, move global params from circuitContext_ to deviceMgr

  // This is done here because the global parameters are needed for the parsing of
  // device lines, but are not known until after pass 1.  This point is immediately
  // before device line parsing on all processors.

  // Map device nodes.
  if (!nodeMap.empty())
  {
    //Stats::StatTop _instantiateStat("Determine Device Nodes");
    //Stats::TimeBlock _instantiateTimer(_instantiateStat);

    std::vector<std::string>::iterator nodeIter = device.getNodeValues().begin();
    std::vector<std::string>::iterator nodeIterEnd = device.getNodeValues().end();
    for (; nodeIter != nodeIterEnd; ++nodeIter)
    {
      nodeMapIter = nodeMap.find( *nodeIter );
      if (nodeMapIter != nodeMap.end())
      {
        // The device node is an external node, map it using nodeMap
        *nodeIter = nodeMapIter->second;
      }
      else if ( *nodeIter != "0" && nodeIter->substr(0,2) != "$G" &&
               !circuitContext_->globalNode( *nodeIter ) )
      {
        // The node is internal, prepend subcircuitPrefix. Note: the
        // ground node and global nodes (0 and $G*) are unchanged.
        *nodeIter = std::string( prefix + ":" + *nodeIter );
      }
    }
  }

  // If the device is not a subcircuit instance add its data to the
  // tables, otherwise, expand the instance.
  if (!device.isSubcircuitInstance())
  {
    //Stats::StatTop _instantiateStat("Process Device Data");
    //Stats::TimeBlock _instantiateTimer(_instantiateStat);

    // Once last check to verify that this is device should be instantiated, 
    // now that the node values have been resolved.
    // NOTE:  It is easier to NOT instantiate the device, than remove it from topology later!
    if (remove_any_redundant_)
    {
      if ( IO::removeDevice(preprocessFilter_, device) )
      {
        return result;
      }
    }

    // Map device name.
    if (prefix != "")
      device.setName(prefix + ":" + device.getInstanceName().getEncodedName());

    // If the device has a model, find it and instantiate it (if it has not
    // already been instantiated). Prepend the model name in the device and
    // the model with the model prefix.  Add the model to the circuit.
    if (device.getModelName() != "")
    {
      //Stats::StatTop _modelStat("Map Model Name");
      //Stats::TimeBlock _modelTimer(_modelStat);

      std::string modelPrefix;
      ParameterBlock* modelPtr;

      bool modelFound = circuitContext_->findModel(device.getModelName(), modelPtr, modelPrefix);
      if (modelFound)
      {
        // Add the model prefix to the device's model name.
        if (modelPrefix != "")
        {
          device.setModelName(modelPrefix + ":" + device.getModelName());
        }
      }
      else
      {
        Report::UserError0() << "Unable to find model " << device.getModelName()
                             << " for device " << device.getInstanceName();
        return false;
      }

      // Add the model to the circuit.
      {
        //Stats::StatTop _modelStat2("Add Model Circuit Block");
        //Stats::TimeBlock _modelTimer2(_modelStat2);
        mainCircuitPtr_->addModel(modelPtr, modelPrefix);
      }
    }

    // If the device is a Bsource, the nodes and sources appearing in
    // the expression must be appropriately prefixed.

    // THIS IS AN AWFUL HACK!  It means that *ANY* device that has
    // "SOLN_DEP" parameters must be hacked in here or they can't be used
    // in a subcircuit!

    if ((device.getNetlistDeviceType() == "B" || device.getNetlistDeviceType() == "S" || device.getNetlistDeviceType() == "C")&& (!nodeMap.empty()))
    {
      Device::Param* givenParameter = 0;
      if (device.getNetlistDeviceType() == "B")
      {
        Device::Param* I_parameter = device.findInstanceParameter( "I" );
        Device::Param* V_parameter = device.findInstanceParameter( "V" );
        if ( I_parameter->given() )
          givenParameter = I_parameter;
        else if ( V_parameter->given() )
          givenParameter = V_parameter;
      }
      else
      {
        Device::Param* C_parameter = 0;
        if (device.getNetlistDeviceType() == "S")
        {
            C_parameter = device.findInstanceParameter( "CONTROL" );
        }
        else if (device.getNetlistDeviceType() == "C")
        {
            C_parameter = device.findInstanceParameter( "C" );
        }

        // Note!  The capacitor could be a semiconductor capacitor, and
        // therefore might not even HAVE a C parameter.  Must guard against
        // that.
        if ( C_parameter->given() )
          givenParameter = C_parameter;
        else
          givenParameter = 0;
      }

      if (givenParameter && givenParameter->getType() == Xyce::Util::EXPR)
      {
        Util::Expression &expression = givenParameter->getValue<Util::Expression>();
        if ( (expression.get_num( XEXP_NODE ) > 0) ||
            (expression.get_num( XEXP_INSTANCE ) > 0) ||
            (expression.get_num( XEXP_LEAD ) > 0) )
        {
          // If the expression has nodes or voltage source instances, get
          // the nodes and map them appropriately or add the subcircuit
          // prefix to them.
          std::vector<std::string> names;
          std::vector<std::string> nodes;
          std::vector<std::string> instances;
          std::vector<std::string> leads;
          if ( expression.get_num( XEXP_NODE ) > 0 )
          {
            expression.get_names( XEXP_NODE, nodes );
            names.insert( names.end(), nodes.begin(), nodes.end() );
          }
          if ( expression.get_num( XEXP_INSTANCE ) > 0 )
          {
            expression.get_names( XEXP_INSTANCE, instances);
            names.insert( names.end(), instances.begin(), instances.end() );
          }
          if ( expression.get_num( XEXP_LEAD ) > 0 )
          {
            expression.get_names( XEXP_LEAD, leads);
            names.insert( names.end(), leads.begin(), leads.end() );
          }

          std::vector<std::string> actualName;
          std::string newName, tmpName;
          unsigned int i;

          nodeMapEnd = nodeMap.end();

          for (i = 0; i < names.size(); ++i )
          {
            nodeMapIter = nodeMap.find(names[i]);
            if (nodeMapIter != nodeMapEnd)
            {
              newName = nodeMapIter->second;
            }
            else if (!(i < nodes.size() && ( names[i].substr(0,2) == "$G" ||
                     circuitContext_->globalNode(names[i]))))
            {
              newName = prefix + ":" + names[i];
            }
            else
            {
              // which leaves the one case, where we're a global node $G...,
              // in which case we just keep the original name.
              newName = names[i];
            }

            // Replace the old node name with the new node name
            // in the expression.
            if (names[i] != newName)
            {
              if (DEBUG_IO)
                Xyce::dout() << "Replacing " << names[i] << " with " << newName << std::endl;

              actualName.push_back(newName);
              tmpName = ";" + newName;
              expression.replace_name( names[i], tmpName );
              if (DEBUG_EXPRESSION) {
                Xyce::dout() << "DistToolBase::instantiateDevice:  After this replacement, get_expression returns "
                             << expression.get_expression() << std::endl
                             << " Parse Tree: " << std::endl;
                expression.dumpParseTree();
              }
            }
          }
          for (i=0 ; i<actualName.size() ; ++i)
          {
            if (DEBUG_IO)
              Xyce::dout() << "Replacing ;"<<actualName[i]<<" with " << actualName[i] << std::endl;

            newName = actualName[i];
            tmpName = ";" + newName;
            expression.replace_name( tmpName, newName);

            if (DEBUG_EXPRESSION) {
              Xyce::dout() << "DistToolBase::instantiateDevice:  After this replacement get_expression returns "
                           << expression.get_expression() << std::endl;
              expression.dumpParseTree();
            }
          }

          // Reset Bsource's expression.
          if (DEBUG_IO)
            Xyce::dout() << "DistToolBase::instantiateDevice:  After all expression handling, get_expression returns "
                         << expression.get_expression() << std::endl;
        }
      }
    }

    if (device.getNetlistDeviceType() == "L" && circuitContext_->haveMutualInductances() )
    {
      handleMutualInductance(device);
    }

    // Add device info to the main circuit device and node tables.
    {
      //Stats::StatTop _addTableStat("Add Table Data");
      //Stats::TimeBlock _addTableTimer(_addTableStat);

      circuitBlock_.addTableData(device);
    }
  }
  else
  {
    result = this->expandSubcircuitInstance(device, libSelect, libInside);
  }

  if (DEBUG_IO)
  {
    device.print();
  }

  return result;
}


//----------------------------------------------------------------------------
// Function       : DistToolBase::handleMutualInductance
// Purpose        : Post-process the mutual inductors in the current circuit,
//                  each inductor must get the list of inductors it is coupled
//                  to, the coupling coeefficient, and model information.
// Special Notes  :
// Scope          :
// Creator        : Rob Hoekstra
// Creation Date  : 08/28/04
//----------------------------------------------------------------------------
bool DistToolBase::handleMutualInductance( DeviceBlock & device )
{
  std::string subcircuitPrefix = circuitContext_->getPrefix();

  //Check for mutual inductance assoc. with this inductor
  std::vector<CircuitContext::MutualInductance> & MIs = circuitContext_->getMutualInductances();

  std::string name = device.getInstanceName().getEncodedName();
  std::string::size_type pos = name.find_last_of(":");
  if (pos != std::string::npos)
    name = name.substr( pos + 1, name.length() - (pos + 1));

  for( unsigned int i = 0; i < MIs.size(); ++i)
  {
    if( MIs[i].inductors.count( name ) )
    {
      // add mutual inductance info to this inductor
      if( MIs[i].inductors.begin()->first == name )
      {
        Device::Param param;

        param.setTag( "FIRSTINDUCTOR" );
        param.setGiven(true);
        param.setVal( 1 );
        device.addInstanceParameter( param );
      }

      if( MIs[i].model != "" )
      {
        std::string modelName(MIs[i].model);

        Device::Param param;

        param.setTag( "NONLINEARCOUPLING" );
        param.setVal( 1 );
        param.setGiven(true);
        device.addInstanceParameter( param );

        std::string modelPrefix;
        ParameterBlock* MI_modelPtr;
        bool modelFound = circuitContext_->
              findModel( modelName, MI_modelPtr, modelPrefix );

        if(modelFound)
        {
          // Check whether the current context is dependent on subcircuit
          // parameters (via a "params:" list on the .subckt line. This
          // will affect the model prefixing.
          if (circuitContext_->hasSubcircuitParams() &&
                MI_modelPtr->hasExpressionValuedParams())
          {
            // If there is a subcircuit parameter dependency, we will
            // assume the worst and add the subcircuit instance prefix
            // to the model prefix. This will insure that model parameters
            // that might depend on subcircuit parameters are assigned to
            // a unique model for each subcircuit instance.
            if( subcircuitPrefix != "" )
              modelPrefix = subcircuitPrefix + ":" + modelPrefix;
          }

          // Add the model prefix to the device's model name.
          if( modelPrefix != "" ) modelName = modelPrefix + ":" + modelName;

          device.setModelName( modelName );
        }
        else
        {
          Report::UserError0() << "Unable to find mutual inductor model " << modelName;
          return false;
        }

        // Add the model to the circuit.
        mainCircuitPtr_->addModel(MI_modelPtr, modelPrefix);
      }
      {
        Device::Param param;

        param.setTag( "COUPLING" );
        param.setVal( MIs[i].coupling );
        param.setGiven(true);
        device.addInstanceParameter( param );
      }

      // add the set of inductors to which it is coupled
      std::map<std::string,double>::iterator iterI = MIs[i].inductors.begin();
      std::map<std::string,double>::iterator  endI = MIs[i].inductors.end();
      for( ; iterI != endI; ++iterI )
      {
        std::string ci = iterI->first;

        if( name != ci )
        {
          Device::Param param;

          param.setTag( "COUPLEDINDUCTOR" );
          if( subcircuitPrefix != "" )
          {
            ci = subcircuitPrefix + ":" + ci;
          }

          param.setVal( ci );
          param.setGiven (true);
          device.addInstanceParameter( param );

          param.setTag("COUPLEDINDUCTANCE");

          if (DEBUG_IO)
            Xyce::dout() << "coupledinductance value from " << name << " to " << ci
                         << " is " << iterI->second << std::endl;

          param.setVal( iterI->second);
          param.setGiven (true);
          device.addInstanceParameter( param );
        }
      }
    }
  }

  return true;
}


} // namespace IO
} // namespace Xyce
