// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Defines the DistToolFlatRoundRobin class.  Distribution tool
//                  buffers and distributes circuit blocks (and related data
//                  such as option blocks, metadata, etc) for/during parsing.
// Special Notes  :
//
// Creator        : Eric Rankin, SNL
//
// Creation Date  : 03/12/2003
//
//
//
//
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_IO_fwd.h>
#include <N_DEV_DeviceMgr.h>
#include <N_ERH_ErrorMgr.h>
#include <N_ERH_Message.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_CmdParse.h>
#include <N_IO_DistToolFlatRoundRobin.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_ParsingHelpers.h>
#include <N_PDS_Comm.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Stats.h>

namespace Xyce {
namespace IO {

//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::DistToolFlatRoundRobin
// Purpose       : ctor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
DistToolFlatRoundRobin::DistToolFlatRoundRobin(
  Parallel::Communicator *                 pdsCommPtr,
  PkgOptionsMgr &                          options_manager,
  CircuitBlock &                           circuit_block,
  std::map<std::string,FileSSFPair>      & ssfMap,
  std::map<std::string, IncludeFileInfo> & iflMap)
  : DistToolBase(circuit_block, ssfMap),
    pdsCommPtr_(pdsCommPtr),
    numProcs_(1),
    currProc_(0),
    optionsManager_(options_manager),
    options_(circuit_block.getOptionsTable()),
    iflMap_(iflMap),
    procDeviceCount_(0),
    blockSize_(0),
    numBlocks_(1),
    devices_(0),
    deviceLinesSent_(0),
    charBufferSize_(250000),
    charBufferPos_(0),
    bufs_(),
    bufSize_(),
    charBuffer_(0),
    circuitContextReady_(false),
    circuitOptionsReady_(false)
{
  numProcs_ = pdsCommPtr_->numProc();
  totalDeviceLinesSent_.resize(numProcs_,0);

  // Set the circuit context and options table.
  setCircuitContext();
  setCircuitOptions();

  // determine how many devices to send to each far proc
  devices_ = (circuit_block.getCircuitContextPtr())->getTotalDeviceCount();
  procDeviceCount_ = devices_ / numProcs_;
  blockSize_ = procDeviceCount_ / numBlocks_;

  // procID == 0 may not be true if we're running in a hierarchical parallel context
  // so check if numProc() == 1 too.
  if (pdsCommPtr_->procID() == 0)
  {
    currProc_ = numProcs_ == 1 ? 0 : 1;
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::~DistToolFlatRoundRobin
// Purpose       : dtor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
DistToolFlatRoundRobin::~DistToolFlatRoundRobin()
{
  delete[] charBuffer_;
}

//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::setCircuitContext
// Purpose       : check buffer size wrt the circuit context size
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolFlatRoundRobin::setCircuitContext()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (pdsCommPtr_->procID() == 0)
    {
      // resize buffer if necessary
      int tmpSize = Xyce::packedByteCount(*circuitContext_) + sizeof( int );

      if ( tmpSize > charBufferSize_ )
      {
        charBufferSize_ = tmpSize;
      }

      // set flag to tx global data
      circuitContextReady_ = true;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::setCircuitOptions
// Purpose       : stage options for tx
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolFlatRoundRobin::setCircuitOptions()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (pdsCommPtr_->procID() == 0)
    {
      // resize buffer if necessary
      int tmpSize = sizeof(int);
      for (std::list<Util::OptionBlock>::const_iterator it = options_.begin(), end = options_.end(); it != end; ++it)
        tmpSize += Xyce::packedByteCount(*it);

      if ( tmpSize > charBufferSize_ )
        charBufferSize_ = tmpSize;

      // set flag to tx global data
      circuitOptionsReady_ = true;
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::circuitDeviceLine
// Purpose       : Send a circuit device line to current proc
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolFlatRoundRobin::circuitDeviceLine(TokenVector & deviceLine )
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if( currProc_ != 0 )
    {
      int size = 0;

      // count line type
      size += sizeof(char);

      // count line size
      size += sizeof(int);

      // count device line
      int deviceLineSize = deviceLine.size();
      for (int i = 0; i < deviceLineSize; ++i)
      {
        size += Xyce::packedByteCount(deviceLine[i]);
      }

      // flush buffer as needed
      send(size);

      // pack line type; "d"evice line
      char lineType = 'd';
      pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );

      // pack device line size
      pdsCommPtr_->pack( &deviceLineSize, 1, charBuffer_, charBufferSize_, charBufferPos_ );

      // pack device line
      for (int i = 0; i < deviceLineSize; ++i)
      {
        Xyce::pack(deviceLine[i],  charBuffer_, charBufferSize_, charBufferPos_, pdsCommPtr_ );
      }

      // increment device line counter
      deviceLinesSent_++;

      // manage proc change if n/p reached
      if (deviceLinesSent_ >= blockSize_)
      {
        totalDeviceLinesSent_[currProc_] += deviceLinesSent_;

        int minus1 = -1;
        int minus2 = -2;

        // flush buffer
        send();

        if (totalDeviceLinesSent_[currProc_] >= procDeviceCount_) 
        {
          // allow this proc to exit receive loop and begin processing
          pdsCommPtr_->send( &minus2, 1, currProc_ );
        }
        else
        {
          // allow this proc to begin processing without exit
          pdsCommPtr_->send( &minus1, 1, currProc_ );
        }

        // reset device line counter
        deviceLinesSent_ = 0;

        // move to next processor
        selectNextProcessor();

        if (currProc_) 
        {
          int length = netlistFilename_.size();
          lineType = 'f';

          // flush buffer as needed
          send(sizeof(char) + sizeof(int) + length);

          // pack the filename
          pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );
          pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
          pdsCommPtr_->pack( netlistFilename_.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );

        }
      }

      return true;
    }
    else
    {
      // Count the lines being processed by Proc 0.
      deviceLinesSent_++;

      if (deviceLinesSent_ >= blockSize_)
      {
        totalDeviceLinesSent_[currProc_] += deviceLinesSent_;

        // reset device line counter
        deviceLinesSent_ = 0;

        // let proc 0 process this line, but then move on to next processor
        // note: this could be the same processor if all other processors have received their allotment.
        selectNextProcessor();
      }

      return false;
    }
  }

  return false;
}

//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::endDeviceLines()
// Purpose       : Make sure other processors have been told that device processing
//                 is ended.  This can be a problem with small device count circuits
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 03/06/06
//-----------------------------------------------------------------------------
void DistToolFlatRoundRobin::endDeviceLines()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    if (currProc_ > 0)
    {
      int minus2 = -2;

      // flush buffer
      send();

      // end parsing on current node
      pdsCommPtr_->send( &minus2, 1, currProc_ );
      ++currProc_;

      // end parsing on remaining nodes
      for ( ; currProc_ < numProcs_ ; ++currProc_)
      {
        pdsCommPtr_->send( &minus2, 1, currProc_ );
      }

      // complete any remaining parser action on node 0
      currProc_ = 0;
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::setFileName
// Purpose       : Change name of netlist file
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
void DistToolFlatRoundRobin::setFileName(std::string const & fileNameIn)
{
  netlistFilename_ = fileNameIn;
  circuitBlock_.setFileName( fileNameIn );

  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    char lineType = 'f';
    int length = netlistFilename_.size();

    // flush buffer as needed
    send(sizeof(char) + sizeof(int) + length);

    pdsCommPtr_->pack( &lineType, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( &length, 1, charBuffer_, charBufferSize_, charBufferPos_ );
    pdsCommPtr_->pack( netlistFilename_.c_str(), length, charBuffer_, charBufferSize_, charBufferPos_ );
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::distributeDevices()
// Purpose       : Distribute the devices according to the strategy.
// Special Notes : The strategy performs a round-robin assignment of devices to
//               : processors to interleave communication and computation.
// Scope         : public
// Creator       : Heidi Thornquist, SNL
// Creation Date : 08/28/2014
//-----------------------------------------------------------------------------
void DistToolFlatRoundRobin::distributeDevices()
{
  Xyce::Parallel::Machine comm = pdsCommPtr_->comm();

  if (DEBUG_IO) {
    Xyce::dout() << "Pass 2 parsing of netlist file and distributing devices: " <<
      circuitBlock_.getNetlistFilename() << std::endl;
  }

  if (Parallel::rank(comm) == 0)
  {
    setFileName(circuitBlock_.getNetlistFilename());

    ssfPtr_ = ssfMap_[netlistFilename_].second;
    ssfPtr_->setLocation( circuitBlock_.getStartPosition() );
    ssfPtr_->setLineNumber( circuitBlock_.getLineStartPosition() );

    // If this is the main circuit, skip over the title line and continue.
    std::ifstream* netlistIn = ssfMap_[netlistFilename_].first;
    std::string title("");
    netlistIn->clear();
    netlistIn->seekg(0, std::ios::beg);
    Xyce::IO::readLine( *netlistIn, title ); 
    ssfPtr_->changeCursorLineNumber( 1 );

    // Instantiate all devices in the current context.
    std::string libSelect, libInside;
    TokenVector line;

    while (getLine(line, libSelect, libInside))
    {
      if (!line.empty() && compare_nocase(line[0].string_.c_str(), ".ends") != 0)
      {
        // parse locally if distool does not distribute
        if( !circuitDeviceLine( line ) )
        {
          handleDeviceLine( line, libSelect, libInside );
        }
      }
    }

    // send MIs if present in top circuit
    if( circuitContext_->haveMutualInductances() )
    {
      // Distribute each YMI?|name device line, end of circuit
      int n = circuitContext_->getNumMILines();
      for( int i = 0; i < n; ++i )
      {
        // parse locally if not distributed to another processor
        if( !circuitDeviceLine( circuitContext_->getMILine( i ) ) )
        {
          handleDeviceLine( circuitContext_->getMILine( i ), libSelect, libInside );
        }
      }
    }

    endDeviceLines();

  }
  else
  {
    // Wait to receive the circuit data from processor 0.
    receiveCircuitData();
  }

  if (DEBUG_IO)
    Xyce::dout() << "Done with pass 2 netlist file parsing and device distribution" << std::endl;

  // Just in case an error was reported in parsing the netlist.
  N_ERH_ErrorMgr::safeBarrier(comm);
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::receiveCircuitData
// Purpose       : receive all data from proc 0 and store it in device buffer
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool DistToolFlatRoundRobin::receiveCircuitData()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    Stats::StatTop _getLineStat("Receive Data");
    Stats::StatTop _processDeviceStat("Process Devices");

    int bsize = 0;
    int num_receives = 0;
    char *currBuffer;

    // The outer loop controls the number of blocks of device buffers that are sent by proc 0.
    while (true)
    {
      // The inner loop controls the number of buffers necessary to send the current block of devices from proc 0.
      while (true)
      {
        Stats::TimeBlock _getLineTimer(_getLineStat);
        pdsCommPtr_->recv( &bsize, 1, 0 );

        // check for halt request, move on to processing buffer.
        if (bsize < 0)
        {
          num_receives++;
          break;
        }

        currBuffer = new char[bsize];
        bufs_.push_back(currBuffer);
        bufSize_.push_back(bsize);

        pdsCommPtr_->recv(currBuffer, bsize, 0);
      }  
    
      bool ok = true;
      {
        std::cout << "Processor " << pdsCommPtr_->procID() << " received buffer number " << num_receives << " in " << bufs_.size() << " blocks. " << std::endl;
        Stats::TimeBlock _processDeviceTimer(_processDeviceStat);
        ok = processDeviceBuffer();
      }
    
      if (bsize == -2)
      {
        return ok;
      }
    }
  }
  else
  {
    return true;
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::send
// Purpose       : send buffer from proc 0 and adjust buffer limits
// Special Notes : size == -1, and no args send(), will force a transmission
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void
DistToolFlatRoundRobin::send(
  int                   size)
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    // transmit buffer if next set of object will fill it completely, or forced
    if ((charBufferPos_ + size >= charBufferSize_) || (size == -1))
    {
      if (DEBUG_DISTRIBUTION)
        dout() << "node " << pdsCommPtr_->procID() << " packed "
               << charBufferPos_ << " bytes into " << charBufferSize_ << " byte buffer"
               << " [data size:  " << size << "]" << std::endl
               << "node " << currProc_ << " was sent "
               << deviceLinesSent_ << " of " << procDeviceCount_ << " devices "
               << (float)deviceLinesSent_ / procDeviceCount_ * 100.0
               << ( size == -1 ? "% complete  [flushed]" : "% complete"  ) << std::endl;

      // tx buffer size
      pdsCommPtr_->send( &charBufferPos_, 1, currProc_ );

      // tx buffer contents
      pdsCommPtr_->send( charBuffer_, charBufferPos_, currProc_ );

      // reset counters
      charBufferPos_ = 0;

      // adjust buffer size if objects larger than existing buffer
      if (size > charBufferSize_)
      {

        // free memory
        // delete [] charBuffer_;

        // set new buffer size
        charBufferSize_ = size;

        // reserve memory; including offset
        charBuffer_ = new char[charBufferSize_ + sizeof(char) + sizeof(int)];

        if (DEBUG_DISTRIBUTION)
          dout() << "node " << pdsCommPtr_->procID()
                 << " resized buffer to " << charBufferSize_
                 << " with offset " << sizeof(char) + sizeof(int) << std::endl;
      }
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::processDeviceBuffer
// Purpose       : process all data in the buffer
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool
DistToolFlatRoundRobin::processDeviceBuffer()
{
  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    int size, bsize, length, pos, i;

    char *currBuffer;
    unsigned int ib;

    // process received device lines, contexts info, and exit calls
    for (ib=0 ; ib<bufs_.size() ; ++ib)
    {
      currBuffer = bufs_[ib];
      bsize = bufSize_[ib];

      // get ready to process buffer
      pos = 0;

      // break apart buffered lines and parse
      while( pos < bsize )
      {
        char lineType = '\0';
        // get the linetype marker that indicates incoming line
        pdsCommPtr_->unpack( currBuffer, bsize, pos, &lineType, 1 );

        // process line according to type
        switch( lineType )
        {
          case 'd': // process a device line
          {
            // unpack the device line
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &size, 1 );
            TokenVector deviceLine( size );

            for( i = 0; i < size; ++i )
            {
              Xyce::unpack(deviceLine[i], currBuffer, bsize, pos, pdsCommPtr_ );
            }

            // hand to circuit block for processing
            std::string libSelect(""), libInside(""); 
            handleDeviceLine( deviceLine, libSelect, libInside );

            break;
          }

          case 'f': // change netlist file name
          {
            pdsCommPtr_->unpack( currBuffer, bsize, pos, &length, 1 );
            netlistFilename_ = std::string( ( currBuffer + pos ), length );
            pos += length;
            setFileName(netlistFilename_);

            break;
          }

          default:  // something went wrong
          {
            Report::DevelFatal().in("DistToolFlatRoundRobin::processDeviceBuffer")
              << "Node " << pdsCommPtr_->procID() << " received invalid message type \"" << lineType;
          }
        }
      }
      delete [] bufs_[ib];
    }

    bufs_.clear();
    bufSize_.clear();
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::broadcastGlobalData
// Purpose       : send bufsize, options, and context to procs
// Special Notes :
// Scope         : private
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool
DistToolFlatRoundRobin::broadcastGlobalData()
{
  // Get the preprocessor filter before broadcasting.
  preprocessFilter_ = circuitBlock_.getPreprocessFilter();

  if (Parallel::is_parallel_run(pdsCommPtr_->comm()))
  {
    // tx buffer size
    pdsCommPtr_->bcast( &charBufferSize_, 1, 0 );

    // check for halt request
    if (charBufferSize_ < 0)
    {
      return false;
    }

    // reserve memory; including offset
    charBuffer_ = new char[charBufferSize_ + sizeof(char) + sizeof(int)];

    // make sure data is ready to tx
    if (numProcs_ > 1)
    {
      int bsize = 0;

      // tx global data
      if (circuitOptionsReady_)
      {
        bsize = Xyce::IO::packCircuitOptions(options_, charBuffer_, charBufferSize_, pdsCommPtr_);
      }

      // broadcast options
      pdsCommPtr_->bcast( &bsize, 1, 0 );
      pdsCommPtr_->bcast( charBuffer_, bsize, 0 );

      // receive global data
      if (!circuitOptionsReady_)
      {
        Xyce::IO::unpackCircuitOptions(options_, charBuffer_, bsize, pdsCommPtr_);
      }

      // tx global data
      if (circuitContextReady_)
      {
        bsize = Xyce::IO::packCircuitContext(circuitContext_, charBuffer_, charBufferSize_, pdsCommPtr_);
      } 
      // broadcast context
      pdsCommPtr_->bcast( &bsize, 1, 0 );
      pdsCommPtr_->bcast( charBuffer_, bsize, 0 );

      if (!circuitContextReady_)
      {
        Xyce::IO::unpackCircuitContext(circuitContext_, charBuffer_, bsize, pdsCommPtr_);
        circuitBlock_.receiveCircuitContext( *circuitContext_ );
      }
    }

    HangingResistor &hanging_resistor = circuitBlock_.getHangingResistor();
  
    // transform booleans into integers and transmit them
    bool netlistcopy = hanging_resistor.getNetlistCopy();
    bool oneterm = hanging_resistor.getOneTerm();
    bool nodcpath = hanging_resistor.getNoDCPath();

    int nlcopyint = 0;
    int otint = 0;
    int nodcint = 0;

    if (netlistcopy)
      nlcopyint = 1;
    if (oneterm)
      otint = 1;
    if (nodcpath)
      nodcint = 1;

    // tx global data
    pdsCommPtr_->bcast(&nlcopyint,1,0);
    pdsCommPtr_->bcast(&otint,1,0);
    pdsCommPtr_->bcast(&nodcint,1,0);

    std::string onetermres(hanging_resistor.getOneTermRes());
    std::string nodcpathres(hanging_resistor.getNoDCPathRes());
    int otreslength = onetermres.size();
    int nodcreslength = nodcpathres.size();

    // tx global data
    pdsCommPtr_->bcast(&otreslength,1,0);
    onetermres.resize(otreslength);
    pdsCommPtr_->bcast(&onetermres[0],otreslength,0);
    pdsCommPtr_->bcast(&nodcreslength,1,0);
    nodcpathres.resize(nodcreslength);
    pdsCommPtr_->bcast(&nodcpathres[0],nodcreslength,0);

    if (nlcopyint == 1)
      hanging_resistor.setNetlistCopy(true);
    if (otint == 1)
      hanging_resistor.setOneTerm(true);
    if (nodcint == 1)
      hanging_resistor.setNoDCPath(true);

    hanging_resistor.setOneTermRes(onetermres);
    hanging_resistor.setNoDCPathRes(nodcpathres);

    N_ERH_ErrorMgr::safeBarrier(pdsCommPtr_->comm());

    // tx preprocess filter
    for (int i=0; i<PreprocessType::NUM_PREPROCESS; i++)
    {
      int preprocint = 0;
      if ( preprocessFilter_[i] )
        preprocint = 1;

      pdsCommPtr_->bcast(&preprocint,1,0);

      if ( preprocint == 1 )
        preprocessFilter_[i] = true;
    }
  }

  if ( preprocessFilter_[PreprocessType::REDUNDANT_C]
       || preprocessFilter_[PreprocessType::REDUNDANT_D]
       || preprocessFilter_[PreprocessType::REDUNDANT_I]
       || preprocessFilter_[PreprocessType::REDUNDANT_L]
       || preprocessFilter_[PreprocessType::REDUNDANT_M]
       || preprocessFilter_[PreprocessType::REDUNDANT_Q]
       || preprocessFilter_[PreprocessType::REDUNDANT_R]
       || preprocessFilter_[PreprocessType::REDUNDANT_V] )
  {
    remove_any_redundant_ = true;
  }

  return true;
}

//--------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::parseIncludeFile
// Purpose       : Jump to include file for 2nd pass
// Special Notes :
// Creator       : Rob Hoekstra
// Creation Date : 08/02/2004
//--------------------------------------------------------------------------
bool DistToolFlatRoundRobin::parseIncludeFile(std::string const& includeFile,
             const std::string &libSelect)
{
  // Save current ssfPtr_ and netlistFilename_.
  SpiceSeparatedFieldTool* oldssfPtr = ssfPtr_;
  std::string old_netlistFilename(netlistFilename_);
  setFileName(includeFile);

  if (DEBUG_IO)
    Xyce::dout() << "Parsing include file Pass 2: " << includeFile << std::endl;

  // Find SSF for file
  if( !ssfMap_.count( includeFile ) )
  {
    Report::UserError() << "Could not find include file SSF " << includeFile;
    return false;
  }
  ssfPtr_ = ssfMap_[includeFile].second;

  //Save current position and move to beginning
  ssfPtr_->setLocation(0);
  ssfPtr_->setLineNumber( 1 );

  TokenVector line;
  std::string empty("");
  while (getLine(line, libSelect, empty)) 
  {
    if (!line.empty() && compare_nocase(line[0].string_.c_str(), ".ends") != 0)
    {
      // parse locally if distool does not distribute
      if( !circuitDeviceLine( line ) )
      {
        handleDeviceLine( line, libSelect, empty );
      }
    }
  }

  // Restore old ssfPtr_ and netlistFilename_.
  ssfPtr_ = oldssfPtr;
  setFileName(old_netlistFilename);

  if (DEBUG_IO)
    Xyce::dout() << "Done with include file Pass 2: " << includeFile << std::endl;

  return true; // Only get here on success.
}

//--------------------------------------------------------------------------
// Function      : DistToolFlatRoundRobin::selectNextProcessor
// Purpose       : Select the next available processor to receive device lines. 
// Special Notes :
// Creator       : Rob Hoekstra
// Creation Date : 08/02/2004
//--------------------------------------------------------------------------
void DistToolFlatRoundRobin::selectNextProcessor()
{
  // If all the other processors have already received their portion of the devices, process
  // the remainder on proc 0;
  if ( (currProc_==0) && (totalDeviceLinesSent_[numProcs_-1] >= procDeviceCount_))
  {
    return;
  }

  // Otherwise, choose the next processor in the loop. 
  currProc_ = (currProc_+1)%numProcs_;
}

} // namespace IO
} // namespace Xyce
