//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : Container class for linear system Jacobian, RHS, Soln Vec,
//                  Error Vec, and Creators N_LAS_QueryUtil and N_PDS_ParMap
//                  are registered to minimize input for matrix and vector
//                  creation
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 06/10/00
//
//
//
//
//-----------------------------------------------------------------------------

#ifndef  Xyce_LAS_System_h
#define  Xyce_LAS_System_h

#include <list>

#include <N_ANP_fwd.h>
#include <N_LAS_fwd.h>

// to eliminate RCP warnings, putting N_PDS_Manager header here.
#include <N_PDS_Manager.h>

// ---------- Forward Declarations ----------

class Epetra_MapColoring;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Class         : System
// Purpose       :
// Special Notes :
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
class System
{

public:
  System();

  ~System();

  // Registration methods for necessary utilities

  // Registers the Parallel Distribution Services (PDS) manager (i.e., sets the
  // pointer.
  bool registerPDSManager(N_PDS_Manager * PDS_Manager)
  { return ( pdsMgr_ = PDS_Manager ); }

  // Registers the Query utility
  bool registerQueryUtil(QueryUtil * LAS_QUtil)
  { return ( lasQueryUtil_ = LAS_QUtil ); }

  // Registers the LAS Builder object
  bool registerBuilder(Builder * builder)
  { return ( lasBuilder_ = builder ); }

  // Registration methods for soln and state vector objects since the TIA or
  // NLS packages will probably own these

  // Registers the pointer to the nonlinear system's Jacobian matrix
  bool registerJacobianMatrix(Matrix * jacMatrixPtr)
  { return ( jacobianMatrixPtr_ = jacMatrixPtr ); }

  // Registers the pointer to the nonlinear system's residual (RHS) vector
  bool registerRHSVector (Vector * rhsVecPtr)
  { return ( rhsVectorPtr_ = rhsVecPtr ); }

  // Registers the pointer to the nonlinear system's limiting (jdxp) vector
  bool registerJDXPVector (Vector * jdxpVecPtr)
  { return ( jdxpVectorPtr_ = jdxpVecPtr ); }

  // dFdxdvp vector, for new-DAE voltage limiting.
  bool registerdFdxdVpVector (Vector * dFdxdVpVectorPtr)
  { return ( dFdxdVpVectorPtr_ = dFdxdVpVectorPtr); }

  // dQdxdvp vector, for new-DAE voltage limiting.
  bool registerdQdxdVpVector (Vector * dQdxdVpVectorPtr)
  { return ( dQdxdVpVectorPtr_ = dQdxdVpVectorPtr) ; }

  // Registers the pointer to the flag solution vector
  bool registerFlagSolVector(Vector * flagSolVecPtr)
  { return ( flagSolVectorPtr_ = flagSolVecPtr ); }

  // Register pointer to device mask
  bool registerDeviceMaskVector(Vector * dmPP)
  { return (deviceMaskVectorPtr_ = dmPP); }
  
  bool registerSolnColoring( Epetra_MapColoring * coloring )
  { return ( solnColoringPtr_ = coloring ); }

  bool registerICColoring( Epetra_MapColoring * coloring )
  { return ( initialConditionColoringPtr_ = coloring ); }

  // Accessor methods for linear system objects

  // Get method for the Jacobian matrix
  Matrix *  getJacobianMatrix() { return jacobianMatrixPtr_; }

  // Get method for the residual (RHS) vector
  Vector *  getRHSVector() { return rhsVectorPtr_; }

  // Get method for the limiting (jdxp) vector
  Vector *  getJDXPVector() { return jdxpVectorPtr_; }

  // dFdxdvp vector, for new-DAE voltage limiting.
  Vector * getdFdxdVpVector () { return  dFdxdVpVectorPtr_; }

  // dQdxdvp vector, for new-DAE voltage limiting.
  Vector * getdQdxdVpVector () { return  dQdxdVpVectorPtr_; }

  // Get method for flag solution vector
  Vector * getFlagSolVector() { return flagSolVectorPtr_; }

  // Get method for the device mask vector
  Vector *  getDeviceMaskVector() { return deviceMaskVectorPtr_; }

  // Registers the Query utility
  QueryUtil * getQueryUtil()
  {
    return lasQueryUtil_;
  }

  // Problem size accessors

  // Get method for the GLOBAL solution vector size (i.e., number of unknowns)
  int getGlobalSolutionSize();

  // Get method for the GLOBAL state vector size (i.e., number of unknowns)
  int getGlobalStateSize();

  // Create residual (RHS) and Jacobian
  bool initializeSystem(); // bool block_analysis_flag);

  // Builder access
  Builder & builder() { return *lasBuilder_; }
  const Builder & builder() const { return *lasBuilder_; }
 
  bool updateExternValsSolnVector(MultiVector * solnVector);
  bool updateExternValsStateVector(MultiVector * stateVector);
  bool updateExternValsStoreVector(MultiVector * storeVector);

  bool localInMatrix(const int & row) const;
  bool localInSolnVector(const int & row) const;
  bool localInStateVector(const int & row) const;

  // Debug call outputs vectors to console
  void debug() const;

private:

  N_PDS_Manager   *  pdsMgr_;

  QueryUtil *  lasQueryUtil_;
  Builder   *  lasBuilder_;

  Matrix    *  jacobianMatrixPtr_;
  Vector    *  rhsVectorPtr_;
  Vector    *  jdxpVectorPtr_;

  Vector    *  dFdxdVpVectorPtr_;
  Vector    *  dQdxdVpVectorPtr_;

  Vector    *  flagSolVectorPtr_;

  Vector    *  deviceMaskVectorPtr_;

  Epetra_MapColoring * solnColoringPtr_;

  Epetra_MapColoring * initialConditionColoringPtr_;
};

} // namespace Linear
} // namespace Xyce

#endif // Xyce_LAS_System_h

