//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
//
// Purpose        : Implementation file for the Iterative linear solver
//                  interface.
//
// Special Notes  :
//
// Creator        : Heidi K. Thornquist, SNL, Electrical & Microsystem Modeling
//
// Creation Date  : 11/11/08
//
//
//
//
//-------------------------------------------------------------------------

// ---------- Standard Includes ----------

#include <Xyce_config.h>

#include <sstream>

// ----------   Xyce Includes   ----------


#include <N_LAS_HBBlockJacobiPrecond.h>
#include <N_LAS_HBBlockJacobiEpetraOperator.h>
#include <N_LAS_HBBuilder.h>
#include <N_LOA_HBLoader.h>
#include <N_MPDE_State.h>

#include <N_LAS_MultiVector.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_Matrix.h>

#include <N_LAS_Problem.h>
#include <N_LAS_Builder.h>
#include <N_LAS_System.h>
#include <N_LAS_SystemHelpers.h>
#include <N_LOA_Loader.h>

#include <N_UTL_Timer.h>
#include <N_UTL_Math.h>

#include <N_ERH_ErrorMgr.h>

#include <N_PDS_SerialComm.h>

#include <Teuchos_RCP.hpp>
#include <Epetra_LinearProblem.h>
#include <Epetra_CrsMatrix.h>
#include <Epetra_MultiVector.h>
#include <Epetra_Import.h>
#include <Epetra_Util.h>
#include <Epetra_Comm.h>
#include <EpetraExt_MatrixMatrix.h>
#include <Amesos.h>

#ifdef Xyce_PARALLEL_MPI
#include <Epetra_MpiComm.h>
#include <mpi.h>
#endif

using Teuchos::RCP;
using Teuchos::rcp;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::HBBlockJacobiPrecond
// Purpose       : Constructor
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
HBBlockJacobiPrecond::HBBlockJacobiPrecond(
  Linear::Builder &builder)
  : isCorrected_(false),
    Preconditioner(),
    builder_(builder)
{
  setDefaultOptions();
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::setDefaultOptions
// Purpose       : resets options
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::setDefaultOptions()
{
  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::setDefaultOption
// Purpose       : resets option
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::setDefaultOption( const std::string & option )
{
  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::setOptions
// Purpose       : sets options and params from modelblock
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::setOptions( const Util::OptionBlock & OB )
{
  // Set the parameters from the list
  Util::ParamList::const_iterator it_tpL = OB.begin();
  Util::ParamList::const_iterator end_tpL = OB.end();
  for (; it_tpL != end_tpL; ++it_tpL)
    {
      this->setParam( *it_tpL );
    }

  // store for restart of solver_
  if( &OB != options_.get() )
    {
      options_ = Teuchos::rcp( &OB, false );
    }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::setParam
// Purpose       : sets options
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::setParam( const Util::Param & param )
{
  std::string uTag = param.uTag();

  // Check if the one-step correction should be used.
  if (uTag=="BLOCK_JACOBI_CORRECTED")
  {
    isCorrected_ = static_cast<bool> (param.getImmutableValue<int>());
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::initGraph
// Purpose       : Initialize the graph using information from the System
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 12/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::initGraph( const Teuchos::RCP<Problem> & problem )
{
  // Generate the graph of each real equivalent form and then generate
  // empty linear systems for each frequency.
  RCP<Matrix> appdQdx = rcp( builder_.createMatrix() );
  RCP<Matrix> appdFdx = rcp( builder_.createMatrix() );

  // Relate the conductance and inductance matrices:
  // G(t) = df/dx(x(t))
  // C(t) = dq/dx(x(t))
  // Compute:  (omega*i*j*C_bar + G_bar)^{-1} for i=0,1,...,M,-M,...,-1
  //           using real equivalent form K_1 = [G_bar -i*omega*C_bar; i*omega*C_bar G_bar]

  // Generate the new real equivalent graph
  RCP<const Epetra_Map> origMap = builder_.getSolutionMap(); 
  int origLocalRows = origMap->NumMyElements();
  int origGlobalRows = origMap->NumGlobalElements();
  int refRows = 2*origLocalRows;
  std::vector<int> rowIdxs( refRows );
  int * origIdxs = origMap->MyGlobalElements();
  for (int i=0; i<origLocalRows; ++i)
  {
    rowIdxs[i] = origIdxs[i];
    rowIdxs[origLocalRows+i] = origIdxs[i] + origGlobalRows;
  }
  epetraMap_ = rcp(new Epetra_Map( -1, refRows, &rowIdxs[0], 0, (appdQdx->epetraObj()).Comm() ) );

  // Count up the number of nonzero entries for the 2x2 block matrix.
  std::vector<int> refNNZs(refRows);
  maxRefNNZs_ = 0;
  for ( int i=0; i<origLocalRows; ++i ) {
    refNNZs[i] = appdQdx->getLocalRowLength(i) + appdFdx->getLocalRowLength(i);
    refNNZs[origLocalRows+i] = refNNZs[i];
    if (refNNZs[i] > maxRefNNZs_) maxRefNNZs_ = refNNZs[i];
  }
  epetraGraph_ = rcp(new Epetra_CrsGraph( Copy, *epetraMap_, &refNNZs[0], true ));

  // Communicate the maximum number of nonzeros for all processors.
  int tmpMaxNNZs = maxRefNNZs_;
  (appdQdx->epetraObj()).Comm().SumAll( &tmpMaxNNZs, &maxRefNNZs_, 1 );

  // Put together the indices for each row and insert them into the graph.
  int tmpNNZs=0, tmpNNZs2=0;
  std::vector<double> tmpCoeffs(maxRefNNZs_);
  std::vector<int> refIdxs(maxRefNNZs_), refIdxs2(maxRefNNZs_);

   for ( int i=0; i<origLocalRows; ++i ) {

    // Get the indices for the first block of the matrix (G_bar)
    appdFdx->getRowCopy( rowIdxs[i], maxRefNNZs_, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

    // Get the indices for the third block of the matrix (C_bar)
    appdQdx->getRowCopy( rowIdxs[i], maxRefNNZs_, tmpNNZs2, &tmpCoeffs[0], &refIdxs2[0] );

    // Insert the indices for the third block into refIdxs, as they are the indices of the second block
    for (int j=0; j<tmpNNZs2; ++j) {
      refIdxs[tmpNNZs+j] = refIdxs2[j]+origGlobalRows;
    }
    epetraGraph_->InsertGlobalIndices( rowIdxs[i], refNNZs[i], &refIdxs[0] );

    // Insert the indices for the first block into refIdxs2, as they are the indices of the fourth block
    for (int j=0; j<tmpNNZs; ++j) {
      refIdxs2[tmpNNZs2+j] = refIdxs[j]+origGlobalRows;
    }
    epetraGraph_->InsertGlobalIndices( rowIdxs[origLocalRows+i], refNNZs[origLocalRows+i], &refIdxs2[0] );
  }
  epetraGraph_->FillComplete();

  // Get the Fourier series information and generate the Epetra_LinearSystems.
  RCP<BlockVector> bXt = hbBuilderPtr_->createTimeDomainBlockVector();
  N_ = bXt->blockCount();
  M_ = (int)((N_-1)/2);

  int numProcs = (appdQdx->epetraObj()).Comm().NumProc();
  int myPID = (appdQdx->epetraObj()).Comm().MyPID();

  if (numProcs > 1)
  {
    // Create the serial maps for all processors.
    // NOTE:  Yes, there must be as many maps/importers/graphs as processors.
    serialEpetraMap_.resize( numProcs );
    serialImporter_.resize( numProcs );
    serialGraph_.resize( numProcs );

    // Determine how many linear systems this processor will manage.
    int nOverP = N_/numProcs;
    beginN_ = myPID*nOverP;
    endN_ = (myPID+1)*nOverP;
    if ( N_%numProcs && myPID==(numProcs-1))
      endN_ += N_%numProcs;

    for (int proc = 0; proc < numProcs; ++proc )
    {
      serialEpetraMap_[proc] = Teuchos::rcp( new Epetra_Map( Epetra_Util::Create_Root_Map( *epetraMap_, proc ) ) );
      serialImporter_[proc] = Teuchos::rcp( new Epetra_Import( *serialEpetraMap_[proc], *epetraMap_ ) ); 
      serialGraph_[proc] = Teuchos::rcp( new Epetra_CrsGraph( Copy, *serialEpetraMap_[proc], 0 ) );
      serialGraph_[proc]->Import( *epetraGraph_, *serialImporter_[proc], Insert );
      serialGraph_[proc]->FillComplete(); 
    }

    // Get the GID list from the serial map, so the split map has the same order of GIDs in parallel.
    int numAll = serialEpetraMap_[myPID]->NumGlobalElements();
    gidList_.resize( numAll );
    for (int i=0; i<numAll; i++)
    {
      gidList_[i] = serialEpetraMap_[myPID]->GID(i);
    }

    // Now split the communicator so that each processor solves their own linear systems.
#ifdef Xyce_PARALLEL_MPI
    MPI_Comm splitComm;
    Epetra_MpiComm mpiComm( dynamic_cast<Epetra_MpiComm &>( const_cast<Epetra_Comm &>( epetraMap_->Comm() ) ) );
    MPI_Comm_split( mpiComm.GetMpiComm(), myPID, myPID, &splitComm); 

    int n=0, maxLen=serialGraph_[myPID]->MaxNumIndices();
    std::vector<int> lIdxs( maxLen ), gIdxs( maxLen );
    Epetra_MpiComm singleComm( splitComm );
    singleMap_ = Teuchos::rcp( new Epetra_Map( numAll, 0, singleComm ) );
    singleGraph_ = Teuchos::rcp( new Epetra_CrsGraph( Copy, *singleMap_, 0 ) );
    for (int i=0; i<numAll; i++)
    {
      serialGraph_[myPID]->ExtractMyRowCopy( i, maxLen, n, &lIdxs[0] );
      for (int j=0; j<n; j++)
      {
        gIdxs[j] = gidList_[ lIdxs[j] ];
      } 
     
      singleGraph_->InsertGlobalIndices( gidList_[i], n, &gIdxs[0] );
    }

    // Now the fill is complete for each graph.
    //for (int proc = 0; proc < numProcs; ++proc )
    //   serialGraph_[proc]->FillComplete(); 
    singleGraph_->FillComplete();
#endif
  } 
  else
  {
    beginN_ = 0;
    endN_ = N_;
  }
  
  Amesos amesosFactory;
  Teuchos::ParameterList params;
  // We are always moving the system to one processor, so the solver should not check inputs to reduce overhead.
  params.set( "TrustMe", true );
  
  // Generate the vectors and matrices for the N_ linear problems to be solved on the diagonal.
  // epetraMatrix_/epetraVector_ : objects having the same distribution as the original system.
  // serialMatrix_/serialVector_ : objects having all rows on a single processor; target for exporting dist objects.
  // singleMatrix_/singleVector_ : objects having all rows on a single processor with a split communicator.
  if (numProcs > 1)
  {
    epetraMatrix_.resize(1);
    epetraMatrix_[0] = Teuchos::rcp( new Epetra_CrsMatrix( Copy, *epetraGraph_ ) );
    epetraMatrix_[0]->FillComplete();
    epetraMatrix_[0]->OptimizeStorage();

    serialVector_ = Teuchos::rcp( new Epetra_MultiVector( *serialEpetraMap_[myPID], 1 ) );

    // Update the importer for performing the matrix importing.
    for (int proc = 0; proc < numProcs; ++proc )
    {
      serialMatrix_ = Teuchos::rcp( new Epetra_CrsMatrix( Copy, *serialGraph_[proc] ) );
      serialImporter_[proc] = Teuchos::rcp( new Epetra_Import( serialMatrix_->RowMap(), epetraMatrix_[0]->RowMap() ) ); 
    }
 
    singleMatrix_.resize(endN_-beginN_);
    epetraProblem_.resize(endN_-beginN_);
    amesosPtr_.resize(endN_-beginN_);

#ifdef Xyce_PARALLEL_MPI 
    singleRHS_ = Teuchos::rcp( new Epetra_MultiVector( *singleMap_, 1 ) );
    singleSoln_ = Teuchos::rcp( new Epetra_MultiVector( *singleMap_, 1 ) );

    for (int i=0; i<endN_-beginN_; ++i) {
      singleMatrix_[i] = rcp( new Epetra_CrsMatrix( Copy, *singleGraph_ ) );
      singleMatrix_[i]->FillComplete();
      singleMatrix_[i]->OptimizeStorage();
      epetraProblem_[i] = rcp( new Epetra_LinearProblem( &*singleMatrix_[i], &*singleRHS_, &*singleSoln_ ) );
      amesosPtr_[i] = rcp( amesosFactory.Create( "Klu", *epetraProblem_[i] ) );
      amesosPtr_[i]->SetParameters( params );
      amesosPtr_[i]->SymbolicFactorization();
    }
#endif
  }
  else 
  {
    epetraRHS_ = Teuchos::rcp( new Epetra_MultiVector( *epetraMap_, 1 ) );
    epetraSoln_ = Teuchos::rcp( new Epetra_MultiVector( *epetraMap_, 1 ) );
    epetraMatrix_.resize(N_);
    epetraProblem_.resize(N_);
    amesosPtr_.resize(N_);

    for (int i=0; i<N_; ++i) {
      epetraMatrix_[i] = rcp( new Epetra_CrsMatrix( Copy, *epetraGraph_ ) );
      epetraMatrix_[i]->FillComplete();
      epetraMatrix_[i]->OptimizeStorage();
      epetraProblem_[i] = rcp( new Epetra_LinearProblem( &*epetraMatrix_[i], &*epetraRHS_, &*epetraSoln_ ) );
      amesosPtr_[i] = rcp( amesosFactory.Create( "Klu", *epetraProblem_[i] ) );
      amesosPtr_[i]->SetParameters( params );
      amesosPtr_[i]->SymbolicFactorization();
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::initValues
// Purpose       : Initialize the values of the Epetra_LinearSystem's
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 12/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::initValues( const Teuchos::RCP<Problem> & problem )
{
  int size_ = freqs_.size();
  int posFreq = (size_-1)/2;

  Teuchos::RCP<Epetra_CrsGraph> linearGraph, nonlinGraph;
  builder_.getSeparatedGraph(linearGraph, nonlinGraph);

  int numProcs = (linearGraph->Comm()).NumProc();
  int myPID = (linearGraph->Comm()).MyPID();

  // Get matrix to contain the sum of all dQdx and dFdx
  // NOTE:  In parallel, this will contain only the nonlinear portion of dQdx and dFdx
  Teuchos::RCP<Matrix> appdQdxSum = rcp( builder_.createMatrix() );
  Teuchos::RCP<Matrix> appdFdxSum = rcp( builder_.createMatrix() );

  // Initialize memory.
  appdQdxSum->put(0.0);
  appdFdxSum->put(0.0);

  // Get the separated stored Jacobian matrices from the HB loader.
  Teuchos::RCP<Matrix>& linAppdQdx = hbLoaderPtr_->getStoreLindQdx();
  Teuchos::RCP<Matrix>& linAppdFdx = hbLoaderPtr_->getStoreLindFdx();

  std::vector<Teuchos::RCP<Matrix> >& vecNLAppdQdx = hbLoaderPtr_->getStoreNLdQdx();
  std::vector<Teuchos::RCP<Matrix> >& vecNLAppdFdx = hbLoaderPtr_->getStoreNLdFdx();

  if (vecNLAppdQdx.size() != 0) 
  {
    // Sum up dQdx and dFdx
    double periodLength = times_[N_] - times_[0];
    for( int i = 0; i < N_; ++i )
    {
      // Compute scale factor.
      double timeStep = times_[i+1] - times_[i];
      double scaleFactor = (timeStep/periodLength);      

      // Add into nonlinear dQdxSum and dFdxSum with scaling
      EpetraExt::MatrixMatrix::Add(vecNLAppdQdx[i]->epetraObj(), false, scaleFactor, 
                                   appdQdxSum->epetraObj(), 1.0);
      EpetraExt::MatrixMatrix::Add(vecNLAppdFdx[i]->epetraObj(), false, scaleFactor, 
                                   appdFdxSum->epetraObj(), 1.0);
    }
  
    // Compute Cdiff and Gdiff of nonlinear matrices, if necessary
    if (isCorrected_)
    {
      // Resize the storage for diffC and diffG.
      diffCMatrix_.resize(N_);
      diffGMatrix_.resize(N_);
      std::vector<Teuchos::RCP<Matrix> > diffC_i(1), diffG_i(1);

      for( int i = 0; i < N_; ++i )
      {
        // Compute C_i - C_avg
        Epetra_CrsMatrix *tmpC_i, *tmpG_i;
        tmpC_i = new Epetra_CrsMatrix( Copy, *nonlinGraph );
        diffC_i[0] = Teuchos::rcp( new Linear::Matrix( tmpC_i, false ) );
        extractValues( *appdQdxSum, diffC_i );
        EpetraExt::MatrixMatrix::Add(vecNLAppdQdx[i]->epetraObj(), false, 1.0, 
                                     *tmpC_i, -1.0);
        diffCMatrix_[i] = Teuchos::rcp( new Linear::Matrix( tmpC_i, true ) );

        // Compute G_i - G_avg
        tmpG_i = new Epetra_CrsMatrix( Copy, *nonlinGraph );
        diffG_i[0] = Teuchos::rcp( new Linear::Matrix( tmpG_i, false ) );
        extractValues( *appdFdxSum, diffG_i );
        EpetraExt::MatrixMatrix::Add(vecNLAppdFdx[i]->epetraObj(), false, 1.0, 
                                     *tmpG_i, -1.0);
        diffGMatrix_[i] = Teuchos::rcp( new Linear::Matrix( tmpG_i, true ) );
      } 
    }
  }

  // Add in the linear contributions, without scaling, if this is a serial run.
  // In parallel, each processor will have a local copy of the linear contributions.
  if(!Teuchos::is_null(linAppdQdx) && (numProcs==1))
  {
    // Add in linear device contributions, without scaling.
    EpetraExt::MatrixMatrix::Add(linAppdQdx->epetraObj(), false, 1.0,
                                 appdQdxSum->epetraObj(), 1.0);
    EpetraExt::MatrixMatrix::Add(linAppdFdx->epetraObj(), false, 1.0,
                                 appdFdxSum->epetraObj(), 1.0);
  }

  // If there is a linear component to the matrix, create a local copy of it for each processor.
  // This is done once and reused throughout the simulation.
  RCP<const Epetra_Map> origMap = builder_.getSolutionMap(); 
  int origLocalRows = origMap->NumMyElements();
  int origGlobalRows = origMap->NumGlobalElements();

  if (!Teuchos::is_null(linAppdQdx) && (lindQdxMatrix_.size()==0) && (numProcs>1)) 
  {
    // If this is parallel, move the linear portion of the matrix so that each processor has a local copy.
    lindFdxMatrix_.resize( numProcs );
    lindQdxMatrix_.resize( numProcs );

    for (int proc = 0; proc < numProcs; ++proc ) 
    {
      // Now create serial objects for the linear portions of the dFdx and dQdx matrices.
      Epetra_Map serialLinMap( Epetra_Util::Create_Root_Map( *origMap, proc ) );
      Epetra_Import serialLinImporter( serialLinMap, *origMap );

      Epetra_CrsGraph serialLinGraph( Copy, serialLinMap, 0 );
      serialLinGraph.Import( linAppdQdx->epetraObj().Graph(), serialLinImporter, Insert );
      serialLinGraph.FillComplete();

      Epetra_CrsMatrix* tmpdFdx = new Epetra_CrsMatrix( Copy, serialLinGraph );
      Epetra_CrsMatrix* tmpdQdx = new Epetra_CrsMatrix( Copy, serialLinGraph );
      Epetra_Import serialLinImporter2( tmpdQdx->RowMap(), linAppdQdx->epetraObj().RowMap() );
      tmpdFdx->Import( linAppdFdx->epetraObj(), serialLinImporter2, Insert );
      tmpdQdx->Import( linAppdQdx->epetraObj(), serialLinImporter2, Insert );
      tmpdFdx->FillComplete();
      tmpdQdx->FillComplete();
      lindFdxMatrix_[proc] = Teuchos::rcp( new Matrix( tmpdFdx, true ) );
      lindQdxMatrix_[proc] = Teuchos::rcp( new Matrix( tmpdQdx, true ) );
    }
  }

  int tmpNNZs=0;
  double cplxCoeff = 0.0;
  std::vector<double> tmpCoeffs(maxRefNNZs_), tmpCoeffs2(2*maxRefNNZs_);
  std::vector<int> refIdxs(maxRefNNZs_), refIdxs2(2*maxRefNNZs_);

  // Get the values for each row of appdFdxSum/appdQdxSum and insert them into the matrix.
  // NOTE:  In serial, this is the complete assembly of the matrix.  In parallel, this is
  // only the assembly of the nonlinear portion of the matrix.

  // Since we are reusing this object, reinitialize it. 
  if (numProcs > 1)
      epetraMatrix_[0]->PutScalar(0.0);

  for ( int nB=0; nB<N_; ++nB ) {

    // Compute the coefficient on the complex matrix [0,1,...,M,-M,...,-1]
    if (nB <= M_)
      cplxCoeff = 2.0 * M_PI * freqs_[posFreq+nB];
    else
      cplxCoeff = 2.0 * M_PI * freqs_[ nB - posFreq -1];

    // Insert the entries for all four blocks of the real-equivalent form
    for ( int i=0; i<origLocalRows; ++i ) {

      // Get the global ID for this row.
      int gid = epetraMap_->GID( i );

      // Load [G_bar -i*omega*C_bar; i*omega*C_bar G_bar]

      // Get the indices for the first block of the matrix (G_bar)
      appdFdxSum->getRowCopy( gid, maxRefNNZs_, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
      
      if (numProcs > 1)
        epetraMatrix_[0]->ReplaceGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
      else
        epetraMatrix_[nB]->ReplaceGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

      // Modify the indices for the fourth block
      for ( int j=0; j<tmpNNZs; ++j ) {
        refIdxs[j] += origGlobalRows;
      }
      if (numProcs > 1)
        epetraMatrix_[0]->ReplaceGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
      else
        epetraMatrix_[nB]->ReplaceGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

      // Add in complex part of REF matrix if non-zero.
      if (cplxCoeff != 0.0) {

        // Get the indices for the first block of the matrix (G_bar)
        appdQdxSum->getRowCopy( gid, maxRefNNZs_, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

        // Insert the second block
        for (int ii=0; ii<tmpNNZs; ++ii) { tmpCoeffs[ii] *= cplxCoeff; }
        if (numProcs > 1)
          epetraMatrix_[0]->ReplaceGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
        else
          epetraMatrix_[nB]->ReplaceGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

        // Insert the third block
        for (int ii=0; ii<tmpNNZs; ++ii) {
          tmpCoeffs[ii] *= -1.0;
          refIdxs[ii] += origGlobalRows;
        }
        if (numProcs > 1)
          epetraMatrix_[0]->ReplaceGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
        else 
          epetraMatrix_[nB]->ReplaceGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
      }
    }

    // Wait for the other processors.
    epetraMatrix_[0]->Comm().Barrier();

    // In parallel, move the current matrix information to a local processor, 
    // the linear contributions will be added in later.
    if (numProcs > 1)
    {
      // Migrate the matrix to one processor.
      int nOverP = N_ / numProcs;
      int proc = nB / nOverP;
      if (proc >= numProcs)
        proc = numProcs-1;

      serialMatrix_ = Teuchos::rcp( new Epetra_CrsMatrix( Copy, *serialGraph_[proc] ) );
      serialMatrix_->Import( *epetraMatrix_[0], *serialImporter_[proc], Insert );
      serialMatrix_->FillComplete();

      // Copy over the matrix entries if this matrix is on my processor.
      if ( nB >= beginN_ && nB < endN_ )
      {
        int idx = nB - beginN_;
        int myRows = serialEpetraMap_[myPID]->NumGlobalElements();
        for (int i=0; i<myRows; i++)
        {
          int gid_i=gidList_[i];
          serialMatrix_->ExtractGlobalRowCopy( gid_i, 2*maxRefNNZs_, tmpNNZs, &tmpCoeffs2[0], &refIdxs2[0] );
          singleMatrix_[idx]->ReplaceGlobalValues( gid_i, tmpNNZs, &tmpCoeffs2[0], &refIdxs2[0] );
        }
      } 
      
      // Wait for the other processors.
      epetraMatrix_[0]->Comm().Barrier();
    }
  } // for ( int nB=0; nB<N_; ++nB ) 


  // In serial, the linear system matrices have now been assembled.
  // In parallel, each processor still needs to add linear contributions to their local linear systems.
  if (numProcs > 1)
  { 
    for ( int nB=beginN_; nB<endN_; ++nB ) 
    {
      int idx = nB - beginN_;

      // Compute the coefficient on the complex matrix [0,1,...,M,-M,...,-1]
      if (nB <= M_)
        cplxCoeff = 2.0 * M_PI * freqs_[posFreq+nB];
      else
        cplxCoeff = 2.0 * M_PI * freqs_[ nB - posFreq -1];

      // Insert the entries for all four blocks of the real-equivalent form
      for ( int i=0; i<origGlobalRows; ++i ) {

        // Get the global ID for this row.
        int gid = i;

        // Load linear [G_bar -i*omega*C_bar; i*omega*C_bar G_bar]

        // Get the indices for the first block of the matrix (G_bar)
        lindFdxMatrix_[myPID]->getRowCopy( gid, maxRefNNZs_, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
    
        singleMatrix_[idx]->SumIntoGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

        // Modify the indices for the fourth block
        for ( int j=0; j<tmpNNZs; ++j ) {
          refIdxs[j] += origGlobalRows;
        }
        singleMatrix_[idx]->SumIntoGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

        // Add in complex part of REF matrix if non-zero.
        if (cplxCoeff != 0.0) {

          // Get the indices for the first block of the matrix (G_bar)
          lindQdxMatrix_[myPID]->getRowCopy( gid, maxRefNNZs_, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

          // Insert the second block
          for (int ii=0; ii<tmpNNZs; ++ii) { tmpCoeffs[ii] *= cplxCoeff; }
          singleMatrix_[idx]->SumIntoGlobalValues( gid+origGlobalRows, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );

          // Insert the third block
          for (int ii=0; ii<tmpNNZs; ++ii) {
            tmpCoeffs[ii] *= -1.0;
            refIdxs[ii] += origGlobalRows;
          }
          singleMatrix_[idx]->SumIntoGlobalValues( gid, tmpNNZs, &tmpCoeffs[0], &refIdxs[0] );
        }
      }
    }

    // Wait for the other processors.
    epetraMatrix_[0]->Comm().Barrier();
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::compute
// Purpose       : Compute a preconditioner M such that M ~= A^{-1}.
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiPrecond::compute()
{
  bool precStatus = true;
  
  // Compute numeric factorization for each block.
  for ( int i=0; i<endN_-beginN_; ++i ) {
    amesosPtr_[i]->NumericFactorization();
  }

  if ( Teuchos::is_null( epetraPrec_ ) )
    epetraPrec_ = blockJacobiOperator( epetraProblem_, amesosPtr_, 
                                       diffCMatrix_, diffGMatrix_, 
                                       hbLoaderPtr_, hbBuilderPtr_, 
                                       freqs_, std::pair<int,int>(beginN_, endN_) );

  if ( Teuchos::is_null( epetraPrec_ ) )
    return false;

  return precStatus;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiPrecond::apply
// Purpose       : Calls the actual preconditioner to apply y = M*x.
// Special Notes :
// Scope         : Public
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 11/11/08
//-----------------------------------------------------------------------------
int HBBlockJacobiPrecond::apply( MultiVector & x, MultiVector & y )
{
  int precStatus = 0;
  // If there is no preconditioner to apply return a nonzero code
  if( Teuchos::is_null(epetraPrec_) )
    precStatus = -1;
  else
    precStatus = epetraPrec_->Apply( x.epetraObj(), y.epetraObj() );

  return precStatus;
}

} // namespace Linear
} // namespace Xyce
