//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2016 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        :
//
// Creator        : Heidi Thornquist, 1437
//
// Creation Date  : 09/04/08
//
//
//
//
//-------------------------------------------------------------------------

// ---------- Standard Includes ----------

// ----------   Xyce Includes   ----------

#include <Xyce_config.h>

#include <N_ERH_ErrorMgr.h>
#include <N_LAS_HBFDJacobianEpetraOperator.h>
#include <N_LAS_HBBuilder.h>
#include <N_LOA_HBLoader.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_Vector.h>
#include <N_PDS_ParMap.h>

// ----------   Trilinos Includes   ----------

#include <Epetra_LinearProblem.h>
#include <Epetra_MultiVector.h>
#include <Epetra_CrsMatrix.h>
#include <Epetra_Vector.h>
#include <Epetra_Comm.h>
#include <Epetra_Map.h>
#include <Amesos_BaseSolver.h>

using Teuchos::RCP;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : matrixFreeEpetraOperator
// Purpose       : non-member constructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 9/4/08
//-----------------------------------------------------------------------------
RCP<HBFDJacobianEpetraOperator> fdJacobianOperator(
    const std::vector<Teuchos::RCP<Epetra_LinearProblem> >& epetraProblems,
    const std::vector<Teuchos::RCP<Amesos_BaseSolver> >& amesosSolvers,
    const Teuchos::RCP<HBBuilder>& hbBuilder,
    const Teuchos::RCP<Loader::HBLoader>& hbLoader,
    const std::vector<double>& timeSteps
    )
{
  RCP<HBFDJacobianEpetraOperator> epetraOperator =
    rcp(new HBFDJacobianEpetraOperator);
  epetraOperator->initialize(epetraProblems,
      amesosSolvers,
      hbBuilder,
      hbLoader,
      timeSteps
      );
  return epetraOperator;
}


//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::HBFDJacobianEpetraOperator
// Purpose       : Constructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
HBFDJacobianEpetraOperator::HBFDJacobianEpetraOperator()
{
  isInitialized_ = false;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::HBFDJacobianEpetraOperator
// Purpose       : Destructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
HBFDJacobianEpetraOperator::~HBFDJacobianEpetraOperator()
{
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::initialize
// Purpose       : Initialization
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
void HBFDJacobianEpetraOperator::initialize(
      const std::vector<Teuchos::RCP<Epetra_LinearProblem> >& epetraProblems,
      const std::vector<Teuchos::RCP<Amesos_BaseSolver> >& amesosSolvers,
      const Teuchos::RCP<HBBuilder>& hbBuilder,
      const Teuchos::RCP<Loader::HBLoader>& hbLoader,
      const std::vector<double>& timeSteps
    )
{
  epetraProblems_ = epetraProblems;
  amesosSolvers_ = amesosSolvers;
  hbBuilder_ = hbBuilder;
  hbLoader_ = hbLoader;
  timeSteps_ = timeSteps;
  N_ = epetraProblems_.size();

  // Initialize work vectors.
  bXtPtr_ = hbBuilder_->createTimeDomainBlockVector();
  xtPtr_ = Teuchos::rcp( new Vector( bXtPtr_->block(0) ) );

  isInitialized_ = true;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::SetUseTranspose
// Purpose       : Define if transpose Apply and ApplyInverse is to be used.
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBFDJacobianEpetraOperator::SetUseTranspose(bool UseTranspose)
{
  // This is not supported for the HB load layers.
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::ApplyInverse
// Purpose       : Apply matrix free preconditioner with Epetra_MultiVectors
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBFDJacobianEpetraOperator::ApplyInverse(
  const Epetra_MultiVector& X,
  Epetra_MultiVector& Y
  ) const
{
  // Convert these to MultiVectors and call the other Apply
  Epetra_MultiVector* Xcopy = new Epetra_MultiVector(X); // This gets deleted by the MultiVector below
  MultiVector las_X(Xcopy, true); // this co-ops the Epetra_MultiVector and uses (and owns) its memory
  MultiVector las_Y(&Y, false);

  int status = ApplyInverse(las_X,las_Y);
  return(status);
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::ApplyInverse
// Purpose       : Apply matrix free preconditioner with MultiVectors
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBFDJacobianEpetraOperator::ApplyInverse(
  const MultiVector& X,
  MultiVector& Y
  ) const
{
  if (!isInitialized_)
  {
    std::string msg = "HBFDJacobianEpetraOperator::ApplyInverse:  I'm not initialized!";
    Report::DevelFatal0() << msg;
  }
  bool status = true;

  // Assume one column in the multivector.
  BlockVector bXf( *(X.getVectorView( 0 )), 2*N_ );

  // Create frequency and time domain work vectors.
  BlockVector bYf( bXf );
  bYf.putScalar( 0.0 );

  // Permute the input vector from the frequency to time domain, since this
  // is a time domain preconditioner.
  hbLoader_->permutedIFT(bXf, &*bXtPtr_);  

  // Now apply preconditioner, which consists of a forward solve since we are using
  // Backward Euler.  The matrix representation of BE is:
  //
  // [ C_1/h+G_1    0        0     ... C_M/h ]
  // [  -C_1/h  C_2/h+G_2    0     ...   0   ]
  // [     0     -C_2/h  C_3/h+G_3 ...   0   ]
  // [ and so on ... ]
  //
  // If the entry in the upper right corner of the matrix is ignored, this is a
  // lower triangular, banded matrix which can be solved with forward solve.

  // Get the separated stored Jacobian matrices from the HB loader.
  Teuchos::RCP<Matrix>& linAppdQdx = hbLoader_->getStoreLindQdx();
  std::vector<Teuchos::RCP<Matrix> >& vecNLAppdQdx = hbLoader_->getStoreNLdQdx();
  
  for (int nB=0; nB<N_; ++nB)
  {
    // Get time step.
    double h = timeSteps_[0];
    if (timeSteps_.size() > 1)
      h = timeSteps_[nB];

    // Get the pointer to the RHS and solution vector, set them for row nB.
    Epetra_MultiVector * nB_RHS = epetraProblems_[nB]->GetRHS();
    Epetra_MultiVector * nB_Soln = epetraProblems_[nB]->GetLHS(); 

    // Subtract off the previous solutions
    // Only need to remove one solution since the matrix is banded.
    if (nB-1 >= 0)
    {
      // Compute nonlinear0 C_Ni-1 * x_i-1.
      if (vecNLAppdQdx.size() != 0)
      {
        (vecNLAppdQdx[nB-1]->epetraObj()).Apply( (bXtPtr_->block(nB-1)).epetraObj(),
                                                 xtPtr_->epetraObj() );
 
        // Compute updated RHS with nonlinear portion:  y_i - C_Ni-1/h * x_i-1 
        (bXtPtr_->block(nB)).addVec( -1.0/h, *xtPtr_ );
      }

      // Compute linear0 C_L-1 * x_i-1.
      if (!Teuchos::is_null(linAppdQdx))
      {
        (linAppdQdx->epetraObj()).Apply( (bXtPtr_->block(nB-1)).epetraObj(),
                                         xtPtr_->epetraObj() );
 
        // Compute updated RHS with linear portion:  y_i - C_L-1/h * x_i-1 
        (bXtPtr_->block(nB)).addVec( -1.0/h, *xtPtr_ );
      }
    }

    // Set the RHS of the linear system
    (*nB_RHS) = (bXtPtr_->block(nB)).epetraObj();

    // Solve the linear system
    amesosSolvers_[nB]->Solve();

    // Copy the solution back to the appropriate bXtPtr block.
    (bXtPtr_->block(nB)).epetraObj() = (*nB_Soln);
  }

  // Permute the output vector from the time domain to the frequency domain
  hbLoader_->permutedFFT(*bXtPtr_, &bYf);

  // Set Y = bYf. 
  Y = bYf;

  if (status)
  {
    return 0;
  }
  else
  {
    return -1;
  }
}
//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::Apply
// Purpose       : Apply inverse of matrix free preconditioner with Epetra_MultiVectors
// Special Notes : Not supported!
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBFDJacobianEpetraOperator::Apply(
  const Epetra_MultiVector& X,
  Epetra_MultiVector& Y
  ) const
{
  std::string msg = "HBFDJacobianEpetraOperator::Apply is not supported!";
  Report::DevelFatal0() <<  msg;
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::Apply
// Purpose       : Apply inverse of matrix free preconditioner with MultiVectors
// Special Notes : Not supported!
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBFDJacobianEpetraOperator::Apply(
  const MultiVector& X,
  MultiVector& Y
  ) const
{
  std::string msg = "HBFDJacobianEpetraOperator::Apply is not supported!";
  Report::DevelFatal0() << msg;
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::NormInf
// Purpose       : Norm Inf of matrix
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
double HBFDJacobianEpetraOperator::NormInf() const
{
  std::string msg = "HBFDJacobianEpetraOperator::NormInf is not supported!";
  Report::DevelFatal0() << msg;
  return -1.0;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::Label
// Purpose       : Label for operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const char * HBFDJacobianEpetraOperator::Label() const
{
  return "Matrix Free Harmonic Balance Finite Difference Preconditioner";
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::UseTranspose
// Purpose       : Query for useTranspose setting
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
bool HBFDJacobianEpetraOperator::UseTranspose() const
{
  // Use Transpose is not supported, so always return false.
  return false;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::HasNormInf
// Purpose       : Query for normInf support
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
bool HBFDJacobianEpetraOperator::HasNormInf() const
{
  // Norm Inf is not supported, so always return false.
  return false;
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::Comm
// Purpose       : Return Epetra_Comm object
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Comm & HBFDJacobianEpetraOperator::Comm() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBFDJacobianEpetraOperator::Comm:  I'm not initialized!";
    Report::DevelFatal0() << msg;
  }
  return((epetraProblems_[0])->GetMatrix()->Comm());
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::OperatorDomainMap
// Purpose       : Return Epetra_Map corresponding to domain of operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Map & HBFDJacobianEpetraOperator::OperatorDomainMap() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBFDJacobianEpetraOperator::OperatorDomainMap:  I'm not initialized!";
    Report::DevelFatal0() << msg;
  }
  return(*(hbBuilder_->getSolutionMap()));
}

//-----------------------------------------------------------------------------
// Function      : HBFDJacobianEpetraOperator::OperatorRangeMap
// Purpose       : Return Epetra_Map corresponding to range of operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Map & HBFDJacobianEpetraOperator::OperatorRangeMap() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBFDJacobianEpetraOperator::OperatorRangeMap:  I'm not initialized!";
    Report::DevelFatal0() << msg;
  }
  return(*(hbBuilder_->getSolutionMap()));
}

} // namespace Linear
} // namespace Xyce
